#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "polarssl/sha2.h"

char sprint_sha256_char(char val) {
	if (val < 10) return val + '0';
	else return val - 10 + 'A';
}
void sprint_sha256(char *str, uint8_t hash[32]) {
	uint32_t i=0;
	for (i=0;i<32;i++) {
		//if ((i)&&((i%8)==0)) {*str = '\n'; str ++;}
		*str = sprint_sha256_char(hash[i]>> 4); str ++;
		*str = sprint_sha256_char(hash[i]&0xF); str ++;
	}
}

int freadwhole(char *name, uint8_t **buf) {
	FILE *fp = 0;
	size_t fl = 0;
	uint8_t *fc = 0;
	fp = fopen(name, "rb");
	if (!fp) return 0;
	fseek(fp, 0, SEEK_END);
	fl = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	fc = (uint8_t*)malloc(fl);
	fread(fc, 1, fl, fp);
	fclose(fp);
	*buf = fc;
	return fl;
}

size_t fsize(char *name) {
	FILE *fp = 0;
	size_t fl = 0;
	fp = fopen(name, "rb");
	if (!fp) return 0;
	fseek(fp, 0, SEEK_END);
	fl = ftell(fp);
	fclose(fp);
	return fl;
}

uint32_t ExtendFileSize(uint32_t ori,uint32_t unit){
	if (ori%unit) return ori-(ori%unit)+unit;
	else return ori;
}

uint32_t findex=0,flcurrent=0; //Section Index, Current Offset
char fname[0x200]={0,}; //Check this filename.
char fchdr[0x200]={0,}; //Exefs Header Contents.

int8_t SubUpdateSection(char *SectionName, char *FileName, char *FileNameAlter, char *FolderName) {
	uint32_t flalt=0,flextended=0;
	uint8_t fx[256] = {0,},fh[32] = {0,}; //Hash in Hex String,Generated SHA256 Hash
	//First FileName
	snprintf(fname,0x200,"%s/%s",FolderName,FileName);
	printf("[INFO]Checking for %s as %s....", FileName, fname);
	flalt=fsize(fname);
	if (flalt) {
		printf("DONE. Found.\n");
	} else {
		if (FileNameAlter) {
			printf("FAIL. Would try %s.\n",FileNameAlter);
			memset(fname,0,0x200);
			snprintf(fname,0x200,"%s/%s",FolderName,FileNameAlter);
			printf("[INFO]Checking for %s as %s....",FileNameAlter,fname);
			flalt=fsize(fname);
			if (flalt) {
				printf("DONE. Found.\n");
				printf("[WARN]Using 3dstool to repack exefs, you should rename %s to %s!\n", FileNameAlter, FileName);
			} else {
				printf("FAIL. No Section %8s found, Step skipped.\n",SectionName);
			}
		} else {
			printf("FAIL.\n");
			printf("[ERR:]Section %8s should be existing as %s in order to rebuild the exefs.\n",SectionName);
			printf("[ERR:]This program can not proceeded, and would exit soon.\n");
		}
	}
	if (flalt) {
		sha2_file(fname, fh, 0);
		sprint_sha256(fx, fh);
		printf("[INFO]%8s hash:%s\n",SectionName,fx);
		memcpy(fchdr+0x0000+0x10*findex,SectionName,8);
		memcpy(fchdr+0x0008+0x10*findex,&flcurrent,4);
		flextended=ExtendFileSize(flalt,0x200);
		if (flextended>flalt)
			printf("[WARN]Enlarge original banner size from 0x%08X to 0x%08X\n", flalt, flextended);
		memcpy(fchdr+0x000C+0x10*findex,&flalt,4);
		memcpy(fchdr+0x01E0-0x20*findex,fh,32);
		findex++;
		flcurrent+=flextended;
		return 1;
	} else {
		printf("[WARN]Section %8s not exists, or Empty. Skipping\n", SectionName);
		return 0;
	}
}

int main(int argc, char** argv) {
	FILE *fhdr=0; //Exefs Header FILE.
	int8_t chkret=0;
	uint32_t flcode=0,fllogo=0,flbanner=0,flicon=0,flalt=0;
	uint8_t *fcfirm=0; //Support rebuild firm exefs.
	uint8_t fx[256] = {0,},fh[32] = {0,}; //Hash in Hex String,Generated SHA256 Hash
	//SectionNames
	char snfirm[0x8]  ={'.','f','i','r','m',0x0,0x0,0x0};
	char sncode[0x8]  ={'.','c','o','d','e',0x0,0x0,0x0};
	char snbanner[0x8]={'b','a','n','n','e','r',0x0,0x0};
	char snicon[0x8]  ={'i','c','o','n',0x0,0x0,0x0,0x0};
	char snlogo[0x8]  ={'l','o','g','o',0x0,0x0,0x0,0x0};
	int i = 0, j = 0; //iterators
	
	printf("[DBG:]argc=%d\n",argc);
	for (i=0;i<argc;i++) printf("[DBG:]argv[%d]=%s\n",i,argv[i]);
	//Test the arguments
	printf("[INFO]Nintendo 3DS Exefs Header Builder.\n");
	if (argc < 2) {
		printf("[SYS:]Please pass enough arguments!\n");
		printf("[HELP]Arguments: <ExefsHeader.bin> <ExefsFolder>\n");
		printf("[HELP]Program would search the ExefsFolder for:\n");
		printf("[HELP]  3dstool: code.bin, banner.bnr, icon.icn, logo.bcma.lz, \n");
		printf("[HELP]  ctrtool: code.bin, banner.bin, icon.bin, logo.bin, \n");
		exit(0);
	}
	memset(fchdr,0,0x200);
	//Loading TMD.
	printf("[INFO]Opening output ExefsHeader %s....", argv[1]);
	fhdr = fopen(argv[1],"wb");
	if (!fhdr) {
		printf("FAIL. Not accessible!\n");
		printf("[ERR:]Since no way for me to write the file, program would exit.\n");
		exit(0);
	}
	printf("DONE.\n");
	
	chkret=SubUpdateSection(snfirm,"firm.bin",0x0, argv[2]);
	if (chkret) {
		flalt=freadwhole(fname,&fcfirm);
		if (fcfirm) {
			printf("[FIRM]Writing ExefsHeader....");
			fwrite(fchdr,1,0x200,fhdr);
			printf("DONE.\n");
			printf("[FIRM]Now writing firm.bin.");
			fwrite(fcfirm,1,flalt,fhdr);
			printf("DONE.\n");
			fclose(fhdr);
			printf("[INFO]Firm exefs built as %s, Program would exit.\n",argv[2]);
			free(fcfirm); exit(0);
		}
	}
	
	chkret=SubUpdateSection(sncode,"code.bin",0x0, argv[2]);
	if (!chkret) {
		printf("[ERR:]Cause None FIRM Exefs must contain a code.bin, action failed.\n");
		printf("[INFO]Program would exit, cause lacking of code.bin\n");
		exit(0);
	}
	
	chkret=SubUpdateSection(snbanner,"banner.bnr","banner.bin", argv[2]);
	chkret=SubUpdateSection(snicon,"icon.icn","icon.bin", argv[2]);
	chkret=SubUpdateSection(snlogo,"logo.bcma.lz","logo.bin", argv[2]);
	
	printf("[INFO]Checking DONE. Exefs would contain %d sections. Now writing file....", findex);
	fwrite(fchdr,1,0x200,fhdr);
	printf("DONE.\n");
	fclose(fhdr);
	
	printf("[INFO]Program proceeded the folder %s", argv[2]);
	exit(0);
}
